/* conf.c
**
** Parsing/Loading configuration from a file
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <ctype.h>

#include "config.h"
#include "prefs.h"
#include "vxerror.h"
#include "tools.h"
#include "mrmacs.h"
#include "dbif.h"

#define LINEMAX 1024

static int num_errs, linenum;

typedef struct {
    char *flagname;
    int flagval;
} flags_t;

static flags_t flags[] = {
    { "admin", ERRORS_ADMIN },
    { "system", ERRORS_SYSTEM },
    { "system+admin", ERRORS_ALL },
    { "admin+system", ERRORS_ALL },
    { NULL, 0 }
};

static flags_t boolflags[] = {
    { "no", 0 },
    { "yes", 1 },
    { NULL, 0 }
};

static flags_t scantype[] = {
    { "regex", 0 },
    { "prefix+empty", 1 },
    { "prefix+prefix", 2 },
    { "prefix", 3 },
    { "prefixonly", 3},
    { NULL, 0 }
};

static flags_t dbtypes[] = {
    { "hash", 0 },
    { "btree", 1 },
    { "recno", 2 },
    { NULL, 0 }
};

static flags_t dbifatflags[] = {
    { "session", DBIFATSESS },
    { "server", DBIFATSERV },
    { NULL, 0 }
};

struct userdb {
    char **name;
    char **module;
} udbref = { &prefs.userdb_name, &prefs.dbif_module } ;

static void set_string (char **svar, char *val);
static void set_dirpath (char **svar, char *val);
/* Sorry, but the following definitions are required in order to make the
** reverse action(s) (vxgetconf) working properly.
*/
static void set_user (char **svar, char *val) { set_string (svar, val); }
static void set_group (char **svar, char *val) { set_string (svar, val); }
/**/
static void set_flag (char *ivar, char *val, flags_t *flags);
static void set_fperm (int *fperm, char *val);
static void set_int (int *ivar, char *val);
static void set_size (long *lvar, char *val);
static void set_module (char **svar, char *val);
static void set_path (char **svar, char *val);
static void set_time (long *ivar, char *val);
static void do_nothing (char **svar, char *val);
static void set_userdb (struct userdb *userdb, char *val);
typedef void (*action_t) (void *tgt, char *val, void *acc);
#define AC(x) ((action_t) (x))
static struct conf {
    char *token;
    enum type { ct_0, ct_f, ct_i, ct_s, ct_db } type;
    action_t action;
    void *tgt;
    void *xtra;
} tokenlist[] = {
    { "conffile", ct_0, AC(do_nothing), &prefs.config_name, NULL },
    { "userdb", ct_db, AC(set_userdb), &udbref, NULL },
    { "dbif_module", ct_s, AC(set_module), &prefs.dbif_module, NULL },
    { "listdir", ct_s, AC(set_dirpath), &prefs.listdir, NULL },
    { "nonlocalto", ct_s, AC(set_string), &prefs.nonlocalto, NULL },
    { "errors", ct_f, AC(set_flag), &prefs.errors, flags },
    { "fperm", ct_i, AC(set_fperm), &prefs.fperm, NULL },
    { "biirm", ct_f, AC(set_flag), &prefs.biirm, boolflags },
    { "maxclients", ct_i, AC(set_int), &prefs.maxclients, NULL },
    { "pop3_service", ct_s, AC(set_string), &prefs.pop3_service, NULL },
    { "smtp_service", ct_s, AC(set_string), &prefs.smtp_service, NULL },
    { "imap_service", ct_s, AC(set_string), &prefs.imap_service, NULL },
    { "vxuser", ct_s, AC(set_user), &prefs.vxuname, NULL },
    { "vxgroup", ct_s, AC(set_group), &prefs.vxgname, NULL },
    { "msg_firstline", ct_s, AC(set_string), &prefs.msg_firstline, NULL },
    { "msg_lastline", ct_s, AC(set_string), &prefs.msg_lastline, NULL },
    { "sysauth", ct_f, AC(set_flag), &prefs.sysauth, boolflags },
    { "check_size", ct_i, AC(set_size), &prefs.check_size, NULL },
    { "check_days", ct_i, AC(set_int), &prefs.check_days, NULL },
    { "mbox_base", ct_s, AC(set_dirpath), &prefs.mbox_base, NULL },
    { "redir_info", ct_s, AC(set_string), &prefs.redir_info, NULL },
    { "dsizecheck", ct_f, AC(set_flag), &prefs.dsizecheck, boolflags },
    { "compression", ct_f, AC(set_flag), &prefs.compression, boolflags },
    { "dbtype", ct_f, AC(set_flag), &prefs.dbtype, dbtypes },
    { "msg_scantype", ct_f, AC(set_flag), &prefs.msg_scantype, scantype },
    { "lcuser", ct_f, AC(set_flag), &prefs.lcuser, boolflags },
    { "lcdomain", ct_f, AC(set_flag), &prefs.lcdomain, boolflags },
    { "syswithidx", ct_f, AC(set_flag), &prefs.syswithidx, boolflags },
    { "maxlevel", ct_i, AC(set_int), &prefs.maxlevel, NULL },
    { "rejverb", ct_f, AC(set_flag), &prefs.rejverb, boolflags },
    { "dracauth", ct_f, AC(set_flag), &prefs.dracauth, boolflags },
    { "bsmtp_nodomain", ct_f, AC(set_flag), &prefs.bsmtp_nodomain, boolflags },
    { "rexecdir", ct_s, AC(set_string), &prefs.rexecdir, NULL },
    { "iu_receives", ct_f, AC(set_flag), &prefs.iu_receives, boolflags },
    { "bsmtp_timeout", ct_i, AC(set_time), &prefs.bsmtp_timeout, NULL },
    { "tmpdir", ct_s, AC(set_dirpath), &prefs.tmpdir, NULL },
    { "error_sender", ct_s, AC(set_string), &prefs.error_sender, NULL },
    { "max_foldersize", ct_i, AC(set_size), &prefs.max_foldersize, NULL },
    { "pop3_autodelete", ct_f, AC(set_flag), &prefs.pop3_autodelete, boolflags},
    { "smtp_maxcmdlen", ct_i, AC(set_int), &prefs.smtp_maxcmdlen, NULL },
    { "sched_cmdfile", ct_s, AC(set_path), &prefs.sched_cmdfile, NULL },
    { "sched_chkt", ct_i, AC(set_int), &prefs.sched_chkt, NULL },
    { "sched_delay", ct_i, AC(set_int), &prefs.sched_delay, NULL },
    { "sched_maxfails", ct_i, AC(set_int), &prefs.sched_maxfails, NULL },
    { "pop3_accept_user",ct_f,AC(set_flag),&prefs.pop3_accept_user,boolflags },
    { "pop3_pass_faildelay",ct_i,AC(set_int),&prefs.pop3_pass_faildelay,NULL },
    { "log_errors", ct_f, AC(set_flag), &prefs.log_errors, boolflags },
    { "loadif_at", ct_f, AC(set_flag), &prefs.loadif_at, dbifatflags },
    { "loglevel", ct_i, AC(set_int), &prefs.loglevel, NULL },
    { "deliver_loglevel", ct_i, AC(set_int), &prefs.deliver_loglevel, NULL },
    { "pop3_loglevel", ct_i, AC(set_int), &prefs.pop3_loglevel, NULL },
    { NULL, ct_0, AC(0), NULL }
};

/* make all strings in the preferences dynamic */

static void init_prefs (void)
{
    static int prefs_saved = 0;
    int ix;
    struct conf *c;
    struct userdb *db;
    if (!prefs_saved) { save_prefs = prefs; prefs_saved = 1; }
    for (ix = 0; noNULL((c = &tokenlist[ix])->token); ++ix) {
	if (c->type == ct_s && noNULL (*((char **) c->tgt))) {
	    *((char **)c->tgt) = gdupl (__FILE__, __LINE__, *((char **)c->tgt));
	} else if (c->type == ct_db) {
	    db = (struct userdb *) c->tgt;
	    unlessnull (*(db->name)) {
		*(db->name) = gdupl (__FILE__, __LINE__, *(db->name));
	    }
	    unlessnull (*(db->module)) {
		*(db->module) = gdupl (__FILE__, __LINE__, *(db->module));
	    }
	}
    }
}

/* unload a dynamically loaded database interface, remove all strings from the
** preferences and reload the compiled defaults.
*/
void conf_reinit (void)
{
    int ix;
    struct conf *c;
    struct userdb *db = &udbref;

    /* unload a dynamically loaded database... */
    if (!lccmp (DBIF_MOD, "_dll", -1)) { dbif_unloadif(); }

    /* remove the database module name and database name */
    unlessnull (*db->name) { gfree (*db->name); *db->name = NULL; }
    unlessnull (*db->module) { gfree (*db->module); *db->module = NULL; }

    /* remove all remaining string values */
    for (ix = 0; noNULL((c = &tokenlist[ix])->token); ++ix) {
	if (c->type == ct_s && noNULL (*((char **)c->tgt))) {
	    gfree (*((char **) c->tgt));
	    *((char **) c->tgt) = NULL;
	}
    }
    prefs = save_prefs;
}

static __inline__ void cfwarn (char *message)
{
    vxerror(0, 0, "%s(%d): WARNING! %s\n", prefs.config_name, linenum, message);
}

static void cferror (char *message)
{
    vxerror (0, 0, "%s(%d): %s\n", prefs.config_name, linenum, message);
    ++num_errs;
}

static void tk_setval (char *tok, char *val)
{
    int ix;
    for (ix = 0; noNULL(tokenlist[ix].token); ++ix) {
	if (!strcmp (tok, tokenlist[ix].token)) {
	    tokenlist[ix].action (tokenlist[ix].tgt, val, tokenlist[ix].xtra);
	    return;
	}
    }
    cferror ("unknown token!\n");
}

char *strskp (char *inp, char *accept)
{
    while (*inp && noNULL(strchr (accept, *inp))) {
	++inp;
    }
    return inp;
}

void getconf (void)
{
    char line[LINEMAX + 1];
    FILE *conf = fopen (prefs.config_name, "r");
    num_errs = 0;
    linenum = 0;
    ifnull (conf) {
	vxerror_at (EX_NOINPUT, errno,
		    __FILE__, __LINE__ - 3,
		    "failed to open configuration file `%s'",
		    prefs.config_name);
    }
    init_prefs();
    while (memset(line, '\0', LINEMAX+1), noNULL(fgets(line, LINEMAX+1, conf)))
    {
	char *p, *q, *v;
	++linenum;
	if (line[strlen (line) - 1] != '\n') {
	    cferror ("line too long!\n");
	}
	p = line + strlen (line);
	if (p != line) {
	    if (*--p == '\r') {
		*p = '\0';		/* MacOS `end of line' */
	    } else if (*p == '\n') {
		*p = '\0';		/* Unix or MSDOS `end of line' */
		if (p != line && *--p == '\r') {
		    *p = '\0';		/* MSDOS `end of line' */
		}
	    }
	}
	p = strskp (line, " \t");
	if (!*p || *p == '#') {
	    continue;
	}
	q = p;
	p = strpbrk (q, " \t");
	ifnull (p) { p = q + strlen (q); } else { *p++ = '\0'; }
	p = strskp (p, " \t");
	v = p;
	tk_setval (q, v);
    }
    fclose (conf);
    if (!strncmp (prefs.userdb_name, "<LIST>", 6)) {
	/* concatenate the `vxworkdir' with the `userdb' in order to get a
	** complete pathname to the userdb
	** i assume that
	** - either `listdir' ends in a `/' or
	** - `userdb_name' is of the form `<LIST>/xyz'.
	*/
	char *newudb, *p = prefs.userdb_name + 6;
	ifnull (newudb = concat (prefs.listdir, p, NULL)) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(concat)");
	}
	gfree (prefs.userdb_name);
	prefs.userdb_name = newudb;
    }
    ifnull (prefs.userdb_name) {
	cferror ("you MUST specify a user database!");
    }
    if (!strcmp (DBIF_MOD, "_dll") && isNULL(prefs.dbif_module)) {
	cferror("you MUST specify the type of the user database you intend to"
		"use!");
    }
    if (num_errs) {
	vxerror_at (EX_SOFTWARE, 0,
		    __FILE__, __LINE__ - 2,
		    "Sorry, but there are syntax errors in the "
		    "configuration file!\n");
    }
}

static void set_string (char **svar, char *val)
{
    *svar = gdupl (__FILE__, __LINE__, val);
}

static void set_userdb (struct userdb *userdb, char *val)
{
    char *dbmod, *p;
    int ix;
    if (noNULL(*(userdb->name))) {
	cferror ("the user database has already been specified");
	return;
    }
    unlessnull (p = strchr (val, ':')) {
	if (p == val) {
	    cferror ("incomplete user database specification - missing db-type");
	    return;
	}
	if (p[1] == '\0') {
	    cferror ("incomplete user database specification - missing database");
	    return;
	}
	unlessnull (*(userdb->module)) {
	    cferror ("you already specified a database module using `dbif_module'");
	    return;
	}
	dbmod = galloc (__FILE__, __LINE__, (p-val) + 9);
	for (ix = 0; ix < (p-val); ++ix) { dbmod[ix] = toupper (val[ix]); }
	strcpy (&dbmod[p-val], "_dbif.so");
	set_module (userdb->module, dbmod);
	gfree (dbmod);
	++p;
    } else {
	p = val;
    }
    set_path (userdb->name, p);
}

static int set_abspath (char **svar, char *val)
{
    if (*val != '/') {
	cferror ("path must be absolute");
	return 1;
    }
    *svar = gdupl (__FILE__, __LINE__, val);
    return 0;
}

static void set_dirpath (char **svar, char *val)
{
    struct stat st;
    char *err, *pfx, *msg;
    size_t sl;
    if (set_abspath (svar, val)) { return; }
    if (stat (val, &st) == -1) {
	err = strerror (errno); pfx = "stat() failed: ";
	sl = strlen (err) + 1 + strlen (pfx);
	msg = gconc (__FILE__, __LINE__, pfx, err, NULL);
	cfwarn (msg);
	gfree (msg);
	return;
    }
    if (!S_ISDIR(st.st_mode)) { cferror ("is no directory!"); }
}

static void set_flag (char *ivar, char *val, flags_t *flags)
{
    int ix;
    for (ix = 0; noNULL(flags[ix].flagname); ++ix) {
	if (!strcmp (flags[ix].flagname, val)) {
	    *ivar = flags[ix].flagval;
	    return;
	}
    }
    cferror ("invalid flag value!");
}

static void set_fperm (int *fperm, char *val)
{
    char *ep;
    int perm = (int) strtol (val, &ep, 8);
    if (noNULL(ep) && (ep == val || *ep)) {
	cferror ("invalid permission mask!");
    }
    *fperm = perm;
}

static void set_int (int *ivar, char *val)
{
    char *ep;
    long res = strtol (val, &ep, 0);
    if (noNULL(ep) && (ep == val || *ep)) {
	cferror ("invalid integer value!");
    }
    *ivar = (int) res;
}

static void set_time (long *ivar, char *val)
{
    char *ep;
    long res = strtol (val, &ep, 10), h, m, s;
    int hset = 0, mset = 0, sset = 0;
    if (ep == val) goto INVALID_TV;
    if (*ep == ':') {
	h = 0; m = res; s = 0;
	val = &ep[1];
	res = strtol (val, &ep, 10);
	if (ep == val || (*ep && *ep != ':')) goto INVALID_TV;
	if (*ep) {
	    h = m; m = res;
	    val = &ep[1];
	    res = strtol (val, &ep, 10);
	    if (ep == val || *ep) goto INVALID_TV;
	    s = res;
	    if (m < 0 || m > 59 || s < 0 || s > 59) goto INVALID_TV;
	    res = (h * 60 + m) * 60 + s;
	} else {
	    s = res;
	    if (m < 0 || m > 59 || s < 0 || s > 59) goto INVALID_TV;
	    res = m * 60 + s;
	}
    } else if (*ep) {
	h = m = s = 0;
	for (;;) {
	    switch (*ep) {
		case 'h':
		    if (hset) goto INVALID_TV;
		    h = 3600 * res; val = &ep[1];
		    break;
		case 'm':
		    if (mset) goto INVALID_TV;
		    m = 60 * res; val = &ep[1];
		    break;
		case 's':
		    if (sset) goto INVALID_TV;
		    s = res; val = &ep[1];
		    break;
		default:
		    goto INVALID_TV;
	    }
	    res = strtol (val, &ep, 10);
	    if (ep == val && !*ep) break;
	}
	if (m < 0 || m > 59 || s < 0 || s > 59) goto INVALID_TV;
	res = (h * 60 + m) * 60 + s;
    }
    *ivar = (int) res;
    return;
INVALID_TV:
    cferror ("invalid time value");
}

static void set_size (long *lvar, char *val)
{
    char *ep;
    long res = strtol (val, &ep, 0);
    if (noNULL(ep) && (ep == val || *ep)) {
	if (noNULL(strchr ("Kk", *ep)) && !ep[1]) {
	    res <<= 10;
	} else if (noNULL(strchr ("Mm", *ep)) && !ep[1]) {
	    res <<= 20;
	} else {
	    cferror ("invalid size value!");
	}
    }
    *lvar = res;
}

static void set_module (char **svar, char *val)
{
    unlessnull (*svar) {
	cferror ("you already specified a database module!");
	return;
    }
    unlessnull (strchr (val, '/')) {
	cferror ("invalid module name (no path allowed)!");
	return;
    }
    *svar = gconc (__FILE__, __LINE__, RXDIR, "/", val, NULL);
}

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

static void set_path (char **svar, char *val)
{
    char *res, *p, *pval, *q, *r;
    struct passwd *pw = getpwuid (geteuid ());
    char cwd[PATH_MAX];
    size_t reslen = strlen (val) + 1, ll;
    ifnull (getcwd (cwd, PATH_MAX)) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "i couldn't get the current working directory");
    }
    pval = val;
    do {
	p = strstr (pval, "<HOME>");
	q = strstr (pval, "<CWD>");
	if (noNULL(p) && (isNULL(q) || (p - q) < 0)) {
	    /* Wenn <HOME> vor <CWD> vorkommt oder nur <HOME> vorhanden ist */
	    reslen += strlen (pw->pw_dir);
	    reslen -= strlen ("<HOME>");
	    pval = p + strlen ("<HOME>");
	} else if (noNULL(q) && (isNULL(p) || (p - q) > 0)) {
	    /* Wenn <CWD> vor <HOME> vorkommt oder nur <CWD> vorhanden ist */
	    reslen += strlen (cwd);
	    reslen -= strlen ("<CWD>");
	    pval = q + strlen ("<CWD>");
	}
    } while (noNULL(p) && noNULL(q));
    res = (char *) galloc (__FILE__, __LINE__, reslen);
    r = res;
    pval = val;
    do {
	p = strstr (pval, "<HOME>");
	q = strstr (pval, "<CWD>");
	if (noNULL(p) && (isNULL(q) || (p - q) < 0)) {
	    memcpy (r, pval, pval - p);
	    r += (pval - p);
	    ll = strlen (pw->pw_dir);
	    memcpy (r, pw->pw_dir, ll);
	    r += ll;
	    *r = '\0';
	    pval = p + strlen ("<HOME>");
	} else if (noNULL(q) && (isNULL(p) || (p - q) > 0)) {
	    memcpy (r, pval, pval - q);
	    r += (pval - q);
	    ll = strlen (cwd);
	    memcpy (r, cwd, ll);
	    r += ll;
	    *r = '\0';
	    pval = q + strlen ("<CWD>");
	}
    } while (noNULL(p) && noNULL(q));
    ll = strlen (pval);
    memcpy (r, pval, ll);
    r[ll] = '\0';
    *svar = res;
}

static void do_nothing (char **svar, char *val)
{
    ;
}

/* Zustzliche Funktion zur Ermittlung der numerischen Werte fr Benutzer-
** und Gruppenidentifikation aus den aus der Konfigurationsdatei eingelesenen
** textuellen Werten.
** Was geschieht hier?
** Fall 1: In der Konfiguration wurde der Name eines existierenden Benutzers
**         bzw. einer existierenden Benutzergruppe angegeben.
** Folge: Aus diesem Namen wird mittels `getpwnam' (bzw. `getgrnam') die
**        zugehrige Benutzeridentifikation (bzw. Gruppenidentifikation)
**        ermittelt und in `prefs' eingetragen.
** Fall 2: In der Konfiguration wurde ein numerischer Wert fr Benutzer- oder
**         Gruppenidentifikation angegeben.
** Folge: Der numerische Wert wird auf seine Gltigkeit hin berprft und da-
**        nach in `prefs' eingetragen. Anschlieend wird versucht, aus diesem
**        Wert (mittels `getpwuid' bzw. `getgrgid') eine symbolische Benutzer-
**        bzw. Gruppenidentifikation zu bestimmen. Ist dies erfolgreich, dann
**        wird diese anstatt des bisherigen Wertes in `prefs' eingetragen;
**        sonst wird eine Warnung ausgegeben, die besagt, da der betreffende
**        Benutzer bzw. die betreffende Gruppe noch nicht existiert; die
**        symbolischen Werte fr Benutzer- bzw. Gruppenidentifikation bleiben
**        im letzteren Fall unverndert.
**
** Der Funktionswert bestimmt sich wie folgt:
**     0 - die Operation war vollstndig erfolgreich,
**     1 - zu einer numerischen Benutzeridentifikation existiert kein Eintrag
**         im System, d.h. es konnte keine symbolische Gruppenidentifikation
**         bestimmt werden.
**     2 - zu einer numerischen Gruppenidentifikation existiert kein Eintrag
**         im System, d.h. es konnte keine symbolische Gruppenidentifikation
**         bestimmt werden.
**     3 - zu keinem der beiden numerischen Werte fr Benutzer- und Gruppen-
**         identifikation existiert ein Eintrag im System.
*/
int get_ug (void)
{
    struct passwd *pw;
    struct group *gr;
    long ival;
    long uimask = -1l ^ (-1l << (8 * sizeof (uid_t)));
    long gimask = -1l ^ (-1l << (8 * sizeof (gid_t)));
    char *p;
    int res = 0;
    unlessnull (pw = getpwnam (prefs.vxuname)) {
	prefs.vxuid = pw->pw_uid;
    } else {
	ival = strtol (prefs.vxuname, &p, 0);
	if ((noNULL(p) && *p) || (ival < 0 || ival > uimask)) {
	    if (abort_on_invug) {
		vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
			    "invalid numerical value for user id!\n");
	    }
	    vxerror(0,0, "WARNING! user `%s' does not exist\n", prefs.vxuname);
	    prefs.vxuid = 0;
	    goto GCNU;
	}
	prefs.vxuid = (uid_t) ival;
	ifnull (pw = getpwuid ((uid_t) ival)) {
	    vxerror (0, 0, "WARNING! No user corresponding to the numerical "
			   "user id found!\n");
	    res |= 1;
	} else {
	    gfree (prefs.vxuname);
	    prefs.vxuname = gdupl (__FILE__, __LINE__, pw->pw_name);
	}
    }
GCNU:
    unlessnull (gr = getgrnam (prefs.vxgname)) {
	prefs.vxgid = gr->gr_gid;
    } else {
	ival = strtol (prefs.vxgname, &p, 0);
	if ((noNULL(p) && *p) || (ival < 0 || ival > gimask)) {
	    if (abort_on_invug) {
		vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
			    "invalid numerical value for group id!\n");
	    }
	    vxerror(0,0, "WARNING! group `%s' does not exist\n", prefs.vxuname);
	    prefs.vxgid = 0;
	    goto GCNG;
	}
	prefs.vxgid = (gid_t) ival;
	ifnull (gr = getgrgid ((gid_t) ival)) {
	    vxerror (0, 0, "WARNING! No group corresponding to the numerical "
			   "group id found!\n");
	    res |= 2;
	} else {
	    gfree (prefs.vxgname);
	    prefs.vxgname = gdupl (__FILE__, __LINE__, gr->gr_name);
	}
    }
GCNG:
    return res;
}


/* Conversion part; required for `vxgetconf' */

typedef char *(*cvproc_t) (void *cval, flags_t *flags);

static char *convmodule (char **_cval);
static char *convstr (char **_cval);
static char *convoct (int *_cval);
static char *convflag (int *_cval, flags_t *flags);
static char *convint (int *_cval);
static char *convuid (char **_cval);
static char *convgid (char **_cval);
static char *convsize (size_t *_cval);
static char *convtime (long *_cval);
static char *convdb (struct userdb *db);

static struct {
    action_t action;
    char *type;
    cvproc_t conv;
} actypes[] = {
    { (action_t) do_nothing, "unsetable string", (cvproc_t) convstr },
    { (action_t) set_string, "string", (cvproc_t) convstr },
    { (action_t) set_dirpath, "string", (cvproc_t) convstr },
    { (action_t) set_userdb, "string", (cvproc_t) convdb },
    { (action_t) set_module, "module", (cvproc_t) convmodule },
    { (action_t) set_path, "path", (cvproc_t) convstr },
    { (action_t) set_fperm, "permissions", (cvproc_t) convoct },
    { (action_t) set_flag, "flag", (cvproc_t) convflag },
    { (action_t) set_int, "integer", (cvproc_t) convint },
    { (action_t) set_user, "user", (cvproc_t) convuid },
    { (action_t) set_group, "group", (cvproc_t) convgid },
    { (action_t) set_size, "size", (cvproc_t) convsize },
    { (action_t) set_time, "time", (cvproc_t) convtime },
    { (action_t) 0, NULL }
};

int confitem (int ix, char **name, char **type, char **value)
{
    int mx, nx;
    for (mx = 0; noNULL(tokenlist[mx].token); ++mx);
    if (ix < mx) {
	for (nx = 0; noNULL(actypes[nx].action); ++nx) {
	    if (tokenlist[ix].action == actypes[nx].action) {
		*name = gdupl (__FILE__, __LINE__, tokenlist[ix].token);
		unlessnull (type) {
		    *type = gdupl (__FILE__, __LINE__, actypes[nx].type);
		}
		*value = actypes[nx].conv (tokenlist[ix].tgt,
					   (flags_t *) tokenlist[ix].xtra);
		return 1;
	    }
	}
    }
    return 0;
}

static char *convdb (struct userdb *db)
{
    char *res, *dbm, *dbn, *pfx = "";
    size_t l, dl;
    ifnull (*(db->module)) {
	/* if the a database module name isn't specified, we probably have
	** a module statically linked against this program.
	*/
	dbm = NULL; l = 0;
    } else {
	/* Otherwise, we have a dynamically loaded module */
	dbm = strrchr (*(db->module), '/');
	ifnull (dbm) { dbm = *(db->module); } else { ++dbm; }
	l = strlen (dbm);
    }
    dbn = *(db->name);
    /* replace a prefix matching `prefs.listdir' by `<LIST>/'. */
    if (noNULL(prefs.listdir)) {
	dl = strlen (prefs.listdir);
	if (!strncmp (dbn, prefs.listdir, dl) && dbn[dl] == '/') {
	    dbn = &dbn[dl+1]; pfx = "<LIST>/";
	}
    }
    if (l > 8 && !strcmp (&dbm[l-8], "_dbif.so")) {
	/* If the name of the database module matches the convention (ending with
	** `_dbif.so') then the `userdb' value should be prefixed with the prefix
	** of the module's name (excluding any directory) - this is then the type
	** of the database and the configuration item `dbif_prefix' should be
	** completely ignored.
	*/
	res = galloc (__FILE__, __LINE__, (l-8)+1 + strlen(pfx) + strlen(dbn) + 1);
	/* construct the result: <dbtype>:<pfx><dbname> */
	memcpy (res, dbm, l-8);
	res[l-8] = ':';
	strcpy (&res[(l-8)+1], pfx);
	strcat (&res[(l-8)+1], dbn);
    } else {
	/* The name of the database module doesn't follow the convention - so
	** it MUST be specified by the `dbif_module' configuration item.
	*/
	res = gconc (__FILE__, __LINE__, pfx, dbn, NULL);
    }
    return res;
}

static char *convmodule (char **_cval)
{
    char *p;
    size_t l;
    ifnull (*_cval) { return (char *) NULL; }
    p = strrchr (*_cval, '/');
    ifnull (p) { p = *_cval; } else { ++p; }
    l = strlen (p);
    if (l > 8 && !strcmp (&p[l-8], "_dbif.so")) {
	return NULL;
    }
    return gdupl (__FILE__, __LINE__, p);
}

static char *convstr (char **_cval)
{
    ifnull (*_cval) { return NULL; }
    return gdupl (__FILE__, __LINE__, *_cval);
}

static char *convoct (int *_cval)
{
    char buf[60], *bufp = buf;
    sprintf (buf, "%04o", *_cval);
    return convstr (&bufp);
}

static char *convflag (int *_cval, flags_t *flags)
{
    int ix, fv = *_cval;
    for (ix = 0; noNULL(flags[ix].flagname); ++ix) {
	if (flags[ix].flagval == fv) {
	    return convstr (&flags[ix].flagname);
	}
    }
    return convstr (&flags[0].flagname);
}

static char *convtime (long *_cval)
{
    char buf[60], *bufp = buf, *p;
    long h = *_cval, m, s;
    s = h % 60; h = h / 60;
    m = h % 60; h = h / 60;
    p = buf; *p = '\0';
    if (h) { sprintf (p, "%ldh", h); p += strlen (p); }
    if (m) { sprintf (p, "%ldm", m); p += strlen (p); }
    if (s) { sprintf (p, "%lds", s); p += strlen (p); }
    if (p == buf) { strcpy (buf, "0"); }
    return convstr (&bufp);
}

static char *convint (int *_cval)
{
    char buf[60], *bufp = buf;
    sprintf (buf, "%d", *_cval);
    return convstr (&bufp);
}

static char *convuid (char **_cval)
{
    struct passwd *pw;
    unlessnull (*_cval) { return convstr (_cval); }
    pw = getpwuid (prefs.vxuid);
    ifnull (pw) { return NULL; }
    return convstr (&pw->pw_name);
}

static char *convgid (char **_cval)
{
    struct group *gr;
    unlessnull (*_cval) { return convstr (_cval); }
    gr = getgrgid (prefs.vxgid);
    ifnull (gr) { return NULL; }
    return convstr (&gr->gr_name);
}

static char *convsize (size_t *_cval)
{
    char buf[60], *bufp = buf;
    size_t sz = *_cval;
    if (!(sz & 0xfffff)) {
	sprintf (buf, "%ldM", (long) (sz >> 20));
    } else if (!(sz & 0x3ff)) {
	sprintf (buf, "%ldK", (long) (sz >> 10));
    } else {
	sprintf (buf, "%ld", (long) (sz));
    }
    return convstr (&bufp);
}
